
#include "quakedef.h"

//cvar_t r_clipwidth = {"r_clipwidth", "640"};
//cvar_t r_clipheight = {"r_clipheight", "480"};
cvar_t r_clipwidth = {"r_clipwidth", "80"};
cvar_t r_clipheight = {"r_clipheight", "60"};

//#define CLIP_MAXPOINTS 65536

//#define WBUFFERSCALE 65536

typedef struct
{
	float w; // inverse depth (1/z)
//	int polynum; // which polygon is at this pixel (used for final rendering pass)
}
clippixel_t;

/*
typedef struct
{
	vec3_t v;
}
clipvert_t;

typedef struct
{
	clipvert_t *firstvert;
	int numverts;
}
clippoly_t;

clipvert_t clipvert[CLIP_MAXPOINTS][3];
*/

typedef struct
{
	float x, w;
}
clipspan_t;

clipspan_t *spanleft, *spanright, *clipspans;

int clipwidth, clipheight;
clippixel_t *clipbuffer;

float r_clip_viewmatrix[3][3], r_clip_viewmulx, r_clip_viewmuly, r_clip_viewcenterx, r_clip_viewcentery, r_clip_nearclipdist, r_clip_nearclipdist2;

void R_Clip_MakeViewMatrix(void)
{
	float pixelaspect, screenaspect, horizontalfieldofview, verticalfieldofview;
	pixelaspect = (float) clipheight / (float) clipwidth * 320 / 240.0;
	horizontalfieldofview = 2.0 * tan (r_refdef.fov_x/360*M_PI);
	screenaspect = clipwidth * pixelaspect / clipheight;
	verticalfieldofview = horizontalfieldofview / screenaspect;
	r_clip_viewcenterx = clipwidth * 0.5 - 0.5;
	r_clip_viewcentery = clipheight * 0.5 - 0.5;
	r_clip_viewmulx = clipwidth / horizontalfieldofview;
	r_clip_viewmuly = r_clip_viewmulx * pixelaspect;
	// this constructs a transposed rotation matrix for the view (transposed matrices do the opposite of their normal behavior)
	VectorCopy (vright, r_clip_viewmatrix[0]);
	VectorNegate (vup, r_clip_viewmatrix[1]);
	VectorCopy (vpn, r_clip_viewmatrix[2]);
	r_clip_nearclipdist = DotProduct(r_origin, vpn) + 4.0f;
	r_clip_nearclipdist2 = r_clip_nearclipdist - 8.0f;
}

void R_Clip_StartFrame(void)
{
	int newwidth, newheight;
	newwidth = bound(80, (int) r_clipwidth.value, vid.width);
	newheight = bound(60, (int) r_clipheight.value, vid.height);
	if (newwidth != clipwidth || newheight != clipheight || !clipbuffer)
	{
		if (clipbuffer)
			qfree(clipbuffer);
		if (clipspans)
			qfree(clipspans);
		clipwidth = newwidth;
		clipheight = newheight;
		clipbuffer = qmalloc(clipwidth * clipheight * sizeof(clippixel_t));
		clipspans = qmalloc(clipheight * sizeof(clipspan_t) * 2);
		spanleft = clipspans;
		spanright = clipspans + clipheight;
	}
	memset(clipbuffer, 0, clipwidth * clipheight * sizeof(clippixel_t));
	R_Clip_MakeViewMatrix();
}

void R_Clip_EndFrame(void)
{
}

void r_clip_start(void)
{
}

void r_clip_shutdown(void)
{
	if (clipbuffer)
		qfree(clipbuffer);
	if (clipspans)
		qfree(clipspans);
	clipbuffer = NULL;
	clipspans = NULL;
	clipwidth = -1;
	clipheight = -1;
}

void r_clip_newmap(void)
{
}

void R_Clip_Init(void)
{
	Cvar_RegisterVariable(&r_clipwidth);
	Cvar_RegisterVariable(&r_clipheight);
	R_RegisterModule("R_Clip", r_clip_start, r_clip_shutdown, r_clip_newmap);
}

static float cv[256*3];
static int currentcv, cvflags[256];

int R_Clip_ClipPolygonToNearClip(float *in, float *out, int inpoints, int stride)
{
	int i, outpoints, prevside, side;
	float *prevpoint, prevdist, dist, dot;

	// begin with the last point, then enter the loop with the first point as current
	prevpoint = (float *) ((byte *)in + stride * (inpoints - 1));
	prevdist = DotProduct(prevpoint, vpn) - r_clip_nearclipdist;
	prevside = prevdist >= 0 ? SIDE_FRONT : SIDE_BACK;
	i = 0;
	outpoints = 0;
	goto begin;
	for (;i < inpoints;i++)
	{
		prevpoint = in;
		prevdist = dist;
		prevside = side;
		(byte *)in += stride;

begin:
		dist = DotProduct(in, vpn) - r_clip_nearclipdist;
		side = dist >= 0 ? SIDE_FRONT : SIDE_BACK;

		if (prevside == SIDE_FRONT)
		{
			VectorCopy(prevpoint, out);
			out += 3;
			outpoints++;
			if (side == SIDE_FRONT)
				continue;
		}
		else if (side == SIDE_BACK)
			continue;

		// generate a split point
		dot = prevdist / (prevdist - dist);
		out[0] = prevpoint[0] + dot * (in[0] - prevpoint[0]);
		out[1] = prevpoint[1] + dot * (in[1] - prevpoint[1]);
		out[2] = prevpoint[2] + dot * (in[2] - prevpoint[2]);
		out += 3;
		outpoints++;
	}

	return outpoints;
}

int R_Clip_ClipTestPolygonToNearClip(float *in, float *out, int inpoints, int stride)
{
	int i, backs = 0, partials = 0;
	float dist;
	for (i = 0;i < inpoints;i++)
	{
		dist = DotProduct(in, vpn);
		if (dist < r_clip_nearclipdist2)
			backs++;
		else if (dist < r_clip_nearclipdist)
			partials++;
		VectorCopy(in, out);
		out += 3;
		(byte *)in += stride;
	}
	if (backs)
	{
		if (backs == inpoints)
			return -1; // completely clipped away
		else
			return 0; // partially clipped
	}
	else if (partials)
		return 0; // partially clipped
	else
		return 1; // completely unclipped
}

static int minyi, maxyi;

int R_Clip_SetupPolygon(int numpoints)
{
	int i, y, y2, minxi, maxxi;
	float *c, minx, maxx, miny, maxy, slopex, x, slopew, w, step, ilength, v[3];
	c = cv;
	for (i = 0;i < numpoints;i++)
	{
		v[0] = c[0] - r_origin[0];
		v[1] = c[1] - r_origin[1];
		v[2] = c[2] - r_origin[2];
		c[0] = r_clip_viewmatrix[0][0] * v[0] + r_clip_viewmatrix[0][1] * v[1] + r_clip_viewmatrix[0][2] * v[2];
		c[1] = r_clip_viewmatrix[1][0] * v[0] + r_clip_viewmatrix[1][1] * v[1] + r_clip_viewmatrix[1][2] * v[2];
		c[2] = r_clip_viewmatrix[2][0] * v[0] + r_clip_viewmatrix[2][1] * v[1] + r_clip_viewmatrix[2][2] * v[2];
		c += 3;
	}
	cv[2] = ilength = 1.0f / cv[2];
	minx = maxx = cv[0] = cv[0] * r_clip_viewmulx * ilength + r_clip_viewcenterx;
	miny = maxy = cv[1] = cv[1] * r_clip_viewmuly * ilength + r_clip_viewcentery;
	if (cv[2] > 1.0f)
		cv[2] = 1.0f;
	c = cv + 3;
	for (i = 1;i < numpoints;i++)
	{
		c[2] = ilength = 1.0f / c[2];
		c[0] = c[0] * r_clip_viewmulx * ilength + r_clip_viewcenterx;
		c[1] = c[1] * r_clip_viewmuly * ilength + r_clip_viewcentery;
		if (cv[2] > 1.0f)
			cv[2] = 1.0f;
		if (miny > c[1])
			miny = c[1];
		else if (maxy < c[1])
			maxy = c[1];
		if (minx > c[0])
			minx = c[0];
		else if (maxx < c[0])
			maxx = c[0];
		c += 3;
	}
	minxi = (int) minx;if (minxi < minx) minxi++;
	maxxi = (int) maxx;if (maxxi < maxx) maxxi++;
	if (minxi < 0)
		minxi = 0;
	if (maxxi > clipwidth)
		maxxi = clipwidth;
	if (minxi >= maxxi)
		return true;
	minyi = (int) miny;if (minxi < minx) minxi++;
	maxyi = (int) maxy;if (maxyi < miny) maxyi++;
	if (minyi < 0)
		minyi = 0;
	if (maxyi > clipheight)
		maxyi = clipheight;
	if (minyi >= maxyi)
		return true;
	// copy the first vert to the vert just after the last to make the rest of the code simpler
	VectorCopy(cv, c);
//	memset(clipspans, 0, sizeof(clipspan_t) * clipheight * 2);
	minyi =  1000000;
	maxyi = -1000000;
	for (i = 0, c = cv;i < numpoints;i++, c += 3)
	{
		if (c[4] > c[1])
		{
			// down edge (right)
			y = (int) c[1];if (y < c[1]) y++;
			y2 = (int) c[4];if (y2 < c[4]) y2++;
			if (y < 0)
				y = 0;
			if (y2 > clipheight)
				y2 = clipheight;
			if (y >= y2)
				continue;
			if (minyi > y)
				minyi = y;
			if (maxyi < y2)
				maxyi = y2;
			step = y - c[1];
			ilength = c[4] - c[1];
			if (ilength >= 1.0f)
				ilength = 1.0f / ilength;
			slopex = (c[3] - c[0]) * ilength;
			x = step * slopex + c[0];
			slopew = (c[5] - c[2]) * ilength;
			w = step * slopew + c[2];
			for (;y < y2;y++)
			{
				spanright[y].x = x;
				spanright[y].w = w;
				x += slopex;
				w += slopew;
			}
		}
		else
		{
			// up edge (left)
			y = (int) c[4];if (y < c[4]) y++;
			y2 = (int) c[1];if (y2 < c[1]) y2++;
			if (y < 0)
				y = 0;
			if (y2 > clipheight)
				y2 = clipheight;
			if (y >= y2)
				continue;
			if (minyi > y)
				minyi = y;
			if (maxyi < y2)
				maxyi = y2;
			step = y - c[4];
			ilength = c[1] - c[4];
			if (ilength >= 1.0f)
				ilength = 1.0f / ilength;
			slopex = (c[0] - c[3]) * ilength;
			x = step * slopex + c[3];
			slopew = (c[2] - c[5]) * ilength;
			w = step * slopew + c[5];
			for (;y < y2;y++)
			{
				spanleft[y].x = x;
				spanleft[y].w = w;
				x += slopex;
				w += slopew;
			}
		}
	}
	if (maxyi < minyi)
		return true;
	// passed setup
	return false;
}

int R_Clip_RenderPolygon(float *inv, int numpoints, int stride)
{
	int y, x, x2, rendered = 0, clipped = 0, valid = 0;
	float slopew, w;
	clippixel_t *cb;
	if (numpoints < 3)
		return false;
	numpoints = R_Clip_ClipPolygonToNearClip(inv, cv, numpoints, stride);
	if (numpoints < 3)
		return false;
	if (R_Clip_SetupPolygon(numpoints))
		return false;
	// test loop: this loop executes until it succeeds in changing a pixel,
	//            then it jumps to the rendering loop, if it never succeeds in
	//            changing a pixel it returns false.
	for (y = minyi;y < maxyi;y++)
	{
		x = (int) spanleft[y].x;if (x < spanleft[y].x) x++;
		x2 = (int) spanright[y].x;if (x2 < spanright[y].x) x2++;
		if (x >= x2)
			continue;
		valid++;
		if (x < 0)
			x = 0;
		if (x2 > clipwidth)
			x2 = clipwidth;
		if (x >= x2)
		{
			clipped++;
			continue;
		}
		rendered++;
		cb = clipbuffer + y * clipwidth;
		slopew = (spanright[y].w - spanleft[y].w) / (spanright[y].x - spanleft[y].x);
		w = (x - spanleft[y].x) * slopew + spanleft[y].w;
		for(;x < x2;x++)
		{
			if (cb[x].w <= w)
				goto rendering;
			w += slopew;
		}
	}
	if (rendered) // pixels were checked but none passed
		return false;
	else if (clipped == valid) // all clipped offscreen
		return false;
	else
		return true; // if not completely clipped away, and all spans were too short, report it as visible

	// rendering loop: this loop simply keeps changing pixels and returns true
	//                 when done.
	for (y = minyi;y < maxyi;y++)
	{
		x = (int) spanleft[y].x;if (x < spanleft[y].x) x++;
		x2 = (int) spanright[y].x;if (x2 < spanright[y].x) x2++;
		if (x < 0)
			x = 0;
		if (x2 > clipwidth)
			x2 = clipwidth;
		if (x >= x2)
			continue;
		cb = clipbuffer + y * clipwidth;
		slopew = (spanright[y].w - spanleft[y].w) / (spanright[y].x - spanleft[y].x);
		w = (x - spanleft[y].x) * slopew + spanleft[y].w;
rendering:
		for(;x < x2;x++)
		{
			if (cb[x].w < w)
				cb[x].w = w;
			w += slopew;
		}
	}
	return true;
}

int R_Clip_TestPolygon(float *inv, int numpoints, int stride)
{
	int y, x, x2, rendered = 0, clipped = 0, valid = 0;
	float slopew, w;
	clippixel_t *cb;
	if (numpoints < 3)
		return false;
	numpoints = R_Clip_ClipPolygonToNearClip(inv, cv, numpoints, stride);
	if (numpoints < 3)
		return false;
	if (R_Clip_SetupPolygon(numpoints))
		return false;
	// test loop: this loop executes until it succeeds in changing a pixel,
	//            then it jumps to the rendering loop, if it never succeeds in
	//            changing a pixel it returns false.
	for (y = minyi;y < maxyi;y++)
	{
		x = (int) spanleft[y].x;if (x < spanleft[y].x) x++;
		x2 = (int) spanright[y].x;if (x2 < spanright[y].x) x2++;
		if (x >= x2)
			continue;
		valid++;
		if (x < 0)
			x = 0;
		if (x2 > clipwidth)
			x2 = clipwidth;
		if (x >= x2)
		{
			clipped++;
			continue;
		}
		rendered++;
		cb = clipbuffer + y * clipwidth;
		slopew = (spanright[y].w - spanleft[y].w) / (spanright[y].x - spanleft[y].x);
		w = (x - spanleft[y].x) * slopew + spanleft[y].w;
		for(;x < x2;x++)
		{
			if (cb[x].w <= w)
				return true;
			w += slopew;
		}
	}
	if (rendered) // pixels were checked but none passed
		return false;
	else if (clipped == valid) // all clipped offscreen
		return false;
	else
		return true; // if not completely clipped away, and all spans were too short, report it as visible
}

int R_Clip_Portal(float *inv, int numpoints, int stride)
{
	int y, x, x2, rendered = 0, clipped = 0, valid = 0;
	float slopew, w, t;
	clippixel_t *cb;
	y = R_Clip_ClipTestPolygonToNearClip(inv, cv, numpoints, stride);
	if (y < 0)
		return false; // completely behind the view
	if (y == 0)
		return true; // partially clipped
	// completely infront
	if (R_Clip_SetupPolygon(numpoints))
		return false;
	for (y = minyi;y < maxyi;y++)
	{
		// portals are considered double sided
		if (spanleft[y].x > spanright[y].x)
		{
			t = spanleft[y].x;
			spanleft[y].x = spanright[y].x;
			spanright[y].x = t;
			t = spanleft[y].w;
			spanleft[y].w = spanright[y].w;
			spanright[y].w = t;
		}
		x = (int) spanleft[y].x;if (x < spanleft[y].x) x++;
		x2 = (int) spanright[y].x;if (x2 < spanright[y].x) x2++;
		if (x >= x2)
			continue;
		valid++;
		if (x < 0)
			x = 0;
		if (x2 > clipwidth)
			x2 = clipwidth;
		if (x >= x2)
		{
			clipped++;
			continue;
		}
		rendered++;
		cb = clipbuffer + y * clipwidth;
		slopew = (spanright[y].w - spanleft[y].w) / (spanright[y].x - spanleft[y].x);
		w = (x - spanleft[y].x) * slopew + spanleft[y].w;
		for(;x < x2;x++)
		{
			if (cb[x].w <= w)
				return true;
			w += slopew;
		}
	}
	if (rendered) // pixels were checked but none passed
		return false;
	else if (clipped == valid) // all clipped offscreen
		return false;
	else
		return true; // if not completely clipped away, and all spans were too short, report it as visible
}

void R_Clip_DisplayBuffer(void)
{
	int i;
	static int firstupload = true;
	byte clipbuffertex[256*256], *b;
	if (!r_render.value)
		return;
	if (clipwidth > 256 || clipheight > 256)
		return;
	glBlendFunc(GL_ONE, GL_ONE);
	glBindTexture(GL_TEXTURE_2D, 8000);
	if (firstupload)
	{
		memset(clipbuffertex, 0, sizeof(clipbuffertex));
		glTexImage2D(GL_TEXTURE_2D, 0, 1, 256, 256, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, clipbuffertex);
	}
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	if (lighthalf)
		glColor3f(0.5, 0.5, 0.5);
	else
		glColor3f(1, 1, 1);
	firstupload = false;
	b = clipbuffertex;
	for (i = 0;i < clipwidth*clipheight;i++)
		*b++ = bound(0, (int) (clipbuffer[i].w * 4096.0f), 255);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, clipwidth, clipheight, GL_LUMINANCE, GL_UNSIGNED_BYTE, clipbuffertex);
	glBegin (GL_QUADS);
	glTexCoord2f (0                 , 0                  );glVertex2f (0        , 0         );
	glTexCoord2f (clipwidth / 256.0f, 0                  );glVertex2f (vid.width, 0         );
	glTexCoord2f (clipwidth / 256.0f, clipheight / 256.0f);glVertex2f (vid.width, vid.height);
	glTexCoord2f (0                 , clipheight / 256.0f);glVertex2f (0        , vid.height);
	glEnd ();
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
//	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

float boxpoints[4*3];

int R_Clip_BoxPolygon(float x1, float y1, float z1, float x2, float y2, float z2, float x3, float y3, float z3, float x4, float y4, float z4)
{
	boxpoints[ 0] = x1;boxpoints[ 1] = y1;boxpoints[ 2] = z1;
	boxpoints[ 3] = x2;boxpoints[ 4] = y2;boxpoints[ 5] = z2;
	boxpoints[ 6] = x3;boxpoints[ 7] = y3;boxpoints[ 8] = z3;
	boxpoints[ 9] = x4;boxpoints[10] = y4;boxpoints[11] = z4;
	return R_Clip_Portal(boxpoints, 4, sizeof(float) * 3);
}

int R_Clip_CullBox(float *a, float *b)
{
	if (R_CullBox(a, b))
		return true;
	if (R_Clip_BoxPolygon(
		a[0], a[1], b[2], 
		b[0], a[1], b[2], 
		b[0], b[1], b[2], 
		a[0], b[1], b[2]))
		return false;
	if (R_Clip_BoxPolygon(
		b[0], a[1], a[2], 
		a[0], a[1], a[2], 
		a[0], b[1], a[2], 
		b[0], b[1], a[2]))
		return false;
	if (R_Clip_BoxPolygon(
		a[0], a[1], a[2], 
		b[0], a[1], a[2], 
		b[0], a[1], b[2], 
		a[0], a[1], b[2]))
		return false;
	if (R_Clip_BoxPolygon(
		b[0], b[1], a[2], 
		a[0], b[1], a[2], 
		a[0], b[1], b[2], 
		b[0], b[1], b[2]))
		return false;
	if (R_Clip_BoxPolygon(
		a[0], a[1], a[2], 
		a[0], b[1], a[2], 
		a[0], b[1], b[2], 
		a[0], a[1], b[2]))
		return false;
	if (R_Clip_BoxPolygon(
		b[0], b[1], a[2], 
		b[0], a[1], a[2], 
		b[0], a[1], b[2], 
		b[0], b[1], b[2]))
		return false;
	return true;
}
